#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <cam_common.h>
#include <cam_ipc.h>
#include <cam_msg.h>
#include <profile.h>

MODULE_DECLARE_MAGIC_STATIC;

#ifdef DEBUG
typedef struct {
	const char *name;
} camd_msg_event_entry_t;

typedef struct {
	const char *name;
} camd_msg_arg_entry_t;

#define ADD_MSG_ID(d, id) d->camd_msg_event_table[id].name=#id
#define ADD_MSG_ARG(d, a) d->camd_msg_arg_table[a].name=#a

static int log_level;

#define DPRINT(l, args...) \
	do { \
		if (l <= log_level) \
			printf(args); \
	} while (0)
#else
#define DPRINT(l, args...) do {} while (0)
#endif /* DEBUG */

#define ERROR(args...) \
	do { \
		snprintf(d->errstr, sizeof(d->errstr), args); \
		DPRINT(0, args); \
	} while (0)

#define STR_ERR_MAX_LEN 128

typedef struct {
#ifdef DEBUG
	camd_msg_event_entry_t camd_msg_event_table[CAMD_MSG_ID_LAST];
	camd_msg_arg_entry_t camd_msg_arg_table[CAMD_MSG_ARG_LAST];
#endif
	char errstr[STR_ERR_MAX_LEN];
	void *h_sipc;
	MODULE_DECLARE_MAGIC_MEMBER;
} camd_msg_data_t;

static camd_msg_data_t camd_msg_singleton;

void *camd_msg_init(void *h_sipc)
{
	camd_msg_data_t *d = &camd_msg_singleton;
	MODULE_REGISTER(d, camd_msg_data_t);

	d->h_sipc = h_sipc;

#ifdef DEBUG
	ADD_MSG_ID(d, CAMD_MSG_ID_N_ILLEGAL);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_POWER);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_PREVIEW);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_FLASH);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_PRE_STILL);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_STILL_CAPTURE);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_VIDEO_CAPTURE);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_TORCH_CURRENT);
	ADD_MSG_ID(d, CAMD_MSG_ID_N_STROBE_CURRENT);
	ADD_MSG_ID(d, CAMD_MSG_ID_Q_IS_PREFLASH_OK);
	ADD_MSG_ID(d, CAMD_MSG_ID_ABORT);

	ADD_MSG_ARG(d, CAMD_MSG_ARG_ILLEGAL);
	ADD_MSG_ARG(d, CAMD_MSG_ARG_STATE_OFF);
	ADD_MSG_ARG(d, CAMD_MSG_ARG_STATE_ON);
	ADD_MSG_ARG(d, CAMD_MSG_ARG_STATE_AUTO);
	ADD_MSG_ARG(d, CAMD_MSG_ARG_NONE);
#endif
	DPRINT(1, "camera IPC messages initialized\n");
	return d;
}

void camd_msg_uninit(void *h)
{
}

#ifdef DEBUG

const char *camd_msg_id2str(void *h, camd_msg_id_t id)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
	return id >= 0 && id < CAMD_MSG_ID_LAST ?
	    d->camd_msg_event_table[id].name : "unknown";

}

const char *camd_msg_arg2str(void *h, camd_msg_arg_t a)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
	return a >= 0 && a < CAMD_MSG_ARG_LAST ?
	    d->camd_msg_arg_table[a].name : "unknown";
}

void camd_msg_contents_disp(void *h, camd_msg_packet_t * p)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);

	DPRINT(1, "client id msg [%s(%d),%s(%d)] payload size %d\n",
	     camd_msg_id2str(d, p->ph.id), p->ph.id,
	     camd_msg_arg2str(d, p->ph.arg), p->ph.arg, p->ph.payload_size);
}

#define SHOW_PCKET_HEADER(p, desc) \
	DPRINT(1, "%s: %s %s::%s, ack %s, payload size %d\n", \
			__FUNCTION__, \
			desc, \
			camd_msg_id2str(d, p->ph.id), \
			camd_msg_arg2str(d, p->ph.arg), \
			p->ph.ack == CAMD_MSG_ACK_REQUEST ? "requested" : "reply", \
			p->ph.payload_size)
#else
#define SHOW_PCKET_HEADER(p, desc) do {} while (0)
#endif

int camd_msg_packet_read_header(void *h, camd_msg_packet_t * p)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
	int ret = sipc_read(d->h_sipc, (void *)p, sizeof(*p)) == sizeof(*p) ? 0 : -1;

	if (ret) 
		ERROR("failed recieving header errno=%d\n", errno);
	else {
		SHOW_PCKET_HEADER(p, "rcvd");
		d->errstr[0] = 0; 
	}

	return ret;
}

int camd_msg_packet_write_header(void *h, camd_msg_packet_t * p)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
	int ret = sipc_write(d->h_sipc, (void *)p, sizeof(*p)) == sizeof(*p) ? 0 : -1;

	if (ret) 
		ERROR("failed sending header errno=%d\n", errno);
	else {
		SHOW_PCKET_HEADER(p, "sent");
		d->errstr[0] = 0; 
	}

	return ret;
}

#ifdef DEBUG
void show_data(void *data, int size)
{
	if (!size) {
		DPRINT(3, "no data to show\n");
		return;
	}

	int i;
	DPRINT(2, "data is %d bytes:\n", size);
	for (i = 0; i < size; i++) {
		DPRINT(2, "0x%x\n", *((char *)data + i));
	}
}
#endif

/* read only the payload of a packet, assuming packet header is in place */
int camd_msg_packet_read_payload(void *h, camd_msg_packet_t *p,
		void *payload_container, int payload_container_size)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
	int ret = -1;

#ifdef DEBUG
	DEBUG_ASSERT(payload_container_size);
	DEBUG_ASSERT(payload_container);
	if (p->ph.payload_size != payload_container_size) {
		ERROR("broken ipc protocol: "
				"p->ph.payload_size=%d payload_container_size=%d "
				"expected to be the same\n",
				p->ph.payload_size, payload_container_size);
		DEBUG_ASSERT(0);
	}
#endif
	if (sipc_read(d->h_sipc, payload_container, payload_container_size) 
			!= p->ph.payload_size) {
		ERROR("%s: read payload failed, expected size %d bytes\n", 
				__FUNCTION__, payload_container_size);
		goto camd_msg_packet_read_payload_exit;
	}
	DPRINT(2, "%s: read payload size %d bytes\n", __FUNCTION__, 
			payload_container_size);
	ret = 0;
	d->errstr[0] = 0; 
camd_msg_packet_read_payload_exit:
	return ret;
}

int camd_msg_packet_write(void *h, camd_msg_packet_t *p)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
	int ret = -1;
	DEBUG_ASSERT(camd_msg_packet_refcount_get(p) > 0);

	if (camd_msg_packet_write_header(d, p))
		goto camd_msg_packet_write_exit;
	/* check payload data integrity */
	DEBUG_ASSERT((p->payload && p->ph.payload_size) || 
			(!p->payload && !p->ph.payload_size));
	/* write payload if necessary */
	if (p->ph.payload_size) {
		DEBUG_ASSERT(p->payload);
		if (sipc_write(d->h_sipc, p->payload, p->ph.payload_size) 
				!= p->ph.payload_size) {
			ERROR("%s: failed writing payload size %d\n", __FUNCTION__,
					p->ph.payload_size);
			goto camd_msg_packet_write_exit;
		}
		else {
			DPRINT(2, "%s: wrote payload size %d\n", __FUNCTION__,
					p->ph.payload_size);
		}
	}
	ret = 0;
	d->errstr[0] = 0; 
camd_msg_packet_write_exit:
	return ret;
}

int camd_msg_packet_write_ack(void *h, camd_msg_packet_t *p, void *payload, 
		int payload_size)
{
#ifdef DEBUG
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
#endif	

	int ret = -1;
	camd_msg_packet_t m;

	camd_msg_packet_init(&m);
	camd_msg_packet_refcount_inc(&m);
	m.ph.id = p->ph.id;
	m.ph.arg = p->ph.arg;
	m.ph.ack = CAMD_MSG_ACK_REPLY;
	m.ph.payload_size = payload_size;
	m.payload = payload;

	if (camd_msg_packet_write(h, &m))
		goto camd_msg_packet_write_ack_exit;
	ret = 0;

camd_msg_packet_write_ack_exit:
	camd_msg_packet_refcount_dec(&m);
	return ret;
}

int camd_msg_packet_write_wait_ack(void *h, camd_msg_packet_t *p)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);

	int ret = -1;
	DEBUG_ASSERT(camd_msg_packet_refcount_get(p) > 0);
	p->ph.ack = CAMD_MSG_ACK_REQUEST;
	if (camd_msg_packet_write(h, p))
		goto camd_msg_packet_write_exit;
	DEBUG_ASSERT(p->ph.ack == CAMD_MSG_ACK_REQUEST);

	DPRINT(1, "%s: sent packet, waiting for ack\n", __FUNCTION__);
	camd_msg_packet_t m;
	camd_msg_packet_refcount_inc(&m);

	/* only header expected in ack */
	if (camd_msg_packet_read_header(d, &m)) {
		camd_msg_packet_refcount_dec(&m);
		goto camd_msg_packet_write_exit;
	}
	DEBUG_ASSERT(m.ph.id == p->ph.id);
	DEBUG_ASSERT(m.ph.ack == CAMD_MSG_ACK_REPLY);
	camd_msg_packet_refcount_dec(&m);
	DPRINT(1, "%s: ack recieved\n", __FUNCTION__);
	ret = 0;
camd_msg_packet_write_exit:
	return ret;
}

void camd_msg_packet_init(camd_msg_packet_t *p)
{
	ZERO(*p);
}

void camd_msg_packet_refcount_inc(camd_msg_packet_t *p)
{
	p->refcount++;
}

void camd_msg_packet_refcount_dec(camd_msg_packet_t *p)
{
	p->refcount--;
	DEBUG_ASSERT(p->refcount >= 0);
}	

int camd_msg_packet_refcount_get(camd_msg_packet_t *p)
{
	return p->refcount;
}

int camd_msg_request_data(void *h, 
		camd_msg_id_t id, 
		camd_msg_arg_t arg, 
		void *data, int size)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);
	int ret = -1;
	camd_msg_packet_t req, reply;
	DEBUG_ASSERT(data && size);

	PROF_SECT_DECL_BEGIN(camd_msg_request_data);

	camd_msg_packet_init(&req);
	camd_msg_packet_init(&reply);
	camd_msg_packet_refcount_inc(&req);
	camd_msg_packet_refcount_inc(&reply);
	req.ph.id = id;
	req.ph.arg = arg;
	req.ph.ack = CAMD_MSG_ACK_REQUEST;

	if (camd_msg_packet_write(d, &req))
		goto read_from_camd_exit;

	if (camd_msg_packet_read_header(d, &reply))
		goto read_from_camd_exit;
	DEBUG_ASSERT(reply.ph.id == id);
	DEBUG_ASSERT(reply.ph.arg == arg);
	DEBUG_ASSERT(reply.ph.ack == CAMD_MSG_ACK_REPLY);

	if (camd_msg_packet_read_payload(d, &reply, data, size))
		goto read_from_camd_exit;

	DPRINT(1, "%s: payload size=%d\n", __FUNCTION__, reply.ph.payload_size);
	ret = 0;
	
read_from_camd_exit:
	camd_msg_packet_refcount_dec(&req);
	camd_msg_packet_refcount_dec(&reply);
	PROF_SECT_END(camd_msg_request_data);
	return ret;
}

const char *camd_msg_get_err_str(void *h)
{
	camd_msg_data_t *d = h;
	MODULE_VERIFY(d);

	return d->errstr;
}
