/* GStreamer
 *
 * Copyright (C) 2001-2002 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *               2006 Edgard Lima <edgard.lima@indt.org.br>
 *
 * gstv4l2src.h: BT8x8/V4L2 source element
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_V4L2SRC_H__
#define __GST_V4L2SRC_H__

#include <poll.h>
#include <semaphore.h>
#include <gstv4l2object.h>
#include <sys/time.h>

#include "cam_msg.h"
#include "cam_ipc.h"
#include "profile.h"

GST_DEBUG_CATEGORY_EXTERN(camsrc_debug);
#define GST_CAT_DEFAULT camsrc_debug

// Flash control
#define FLASH_OFF		0
#define FLASH_ON		1
#define FLASH_AUTO		2

// Caputre errors
#define CAPTURE_SUCCESS         0
#define CAPTURE_PENDING         1
#define CAPTURE_ERR_FILE        2	// error in file saving
#define CAPTURE_ERR_DATA        3	// gstreamer error and OMX errors
#define CAPTURE_ERR_SYSTEM      4	// V4l2 driver or system related error
					// e.g. malloc failed.
#define CAPTURE_START		5	// preview stopped, exposure started

/* uncomment to enable store of preview frames, just before still capture */
//#define STORE_PRE_STILL_FRAMES
//#define STORE_STILL_FRAMES

#if defined(STORE_PRE_STILL_FRAMES) || defined(STORE_STILL_FRAMES)
#define ENABLE_FRAME_STORE
#endif					

G_BEGIN_DECLS
#define GST_TYPE_V4L2SRC \
  (gst_v4l2src_get_type())
#define GST_V4L2SRC(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_V4L2SRC,GstV4l2Src))
#define GST_V4L2SRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_V4L2SRC,GstV4l2SrcClass))
#define GST_IS_V4L2SRC(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_V4L2SRC))
#define GST_IS_V4L2SRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_V4L2SRC))

typedef struct _GstV4l2Buffer GstV4l2Buffer;
typedef struct _GstV4l2Src GstV4l2Src;
typedef struct _GstV4l2SrcClass GstV4l2SrcClass;

/* global info */

struct _GstV4l2Buffer {
	GstBuffer buffer;
	struct v4l2_buffer vbuffer;
	GstV4l2Src *v4l2src;
};

#ifdef ENABLE_FRAME_STORE
typedef struct {
	int session_num;
	int frame_num;
	char *desc;
	char *filename;
	gboolean active;
} frame_store_seq_t;
#endif

/* poll events */
enum {
	PE_V4L2 = 0,
	PE_IPC,
	PE_NOTIFY,
	PE_LAST
};

#ifdef DEBUG
//#define SHOW_FPS
//#define SHOW_FRAMES_QDQ
//#define SHOW_TORCH_CONVERGE_SEQ
//#define SHOW_TIMESTAMPS_DELTA
#define WARN_FRAME_STARVATION
#endif

#define NUM_OVERLAY_BUFFERS 20

/**
 * GstV4l2Src:
 * @pushsrc: parent #GstPushSrc.
 *
 * Opaque object.
 */
struct _GstV4l2Src {
	GstPushSrc pushsrc;

	/* < private > */
	GstV4l2Object *v4l2object;

	/* still image src pad */
	GstPad *src_still_pad;
	GMutex *mutex;
	gboolean flash_required;
	struct v4l2_format still_format;
	guint32 still_duration;
	GstV4l2Buffer *still_capture_buffer;
	GstClockTime still_capture_start_ts;
	GstClockTime last_video_buffer_ts;
	struct timeval tv_start;
	guint fps_n;
	guint fps_d;
	gint width;
	gint height;
	guint32 pixelformat;

	/* internal lists */
	GSList *formats;	/* list of available capture formats */

	/* buffers */
	struct {
		GstBuffer *ptr;
		gboolean queued;
	} video_buf[NUM_OVERLAY_BUFFERS];

	struct v4l2_buffer vbuffer;

	int num_queued_buffers;
	int dq_frame_count;
#ifdef SHOW_FPS
	int frame_count_last;
	GstClockTime frame_count_ts;
#endif
	guint32 frame_byte_size;

	/* flashlight related */
	guint32 flash_ctrl;
	guint32 current_flash_ctrl;
	enum {
		FCM_PREVIEW_1ST_FRAME = 0,
		FCM_PREVIEW,
		FCM_PREVIEW_W_TORCH,
		FCM_STILL_CAPTURE,
	} frame_create_mode;

	guint64 offset;
	gboolean init_complete;
	sem_t still_capture_sem;
	GMutex *post_still_hold_frame_lock;
	GMutex *still_buffer_lock;
	gboolean pausing_after_still;
	GCond *still_hold_frame_cond;
	gboolean still_buffer_ready;
	gboolean relinquish_still_buffer;

	/* camd related members */
	GMutex *camd_comm_lock;
	void *h_camd_msg;
	void *h_sipc;
	camd_msg_post_still_data_t psd;

	/* poll related */
	struct pollfd pem[PE_LAST]; /* poll event monitor */
	gint num_monitored_poll_events;
	struct {
		int fd;
		int poll_idx;
	} event_info[PE_LAST];
	GMutex *notification_lock;

#ifdef ENABLE_FRAME_STORE 
	frame_store_seq_t *preview_fh;
#endif
#ifdef ENABLE_FRAME_STORE 
	frame_store_seq_t *still_fh;
#endif
#ifdef YUVS_ENABLE
	struct {
		pthread_t thread_id;
		gboolean shutdown;
		sem_t sem;
		GMutex *lock;
		GQueue *queue;
	} yuvs;
#endif
	gint num_discard_video_frames;
};

struct _GstV4l2SrcClass {
	GstPushSrcClass parent_class;

	GList *v4l2_class_devices;
};

GType gst_v4l2src_get_type(void);

G_END_DECLS

#ifdef SHOW_EXTRA_LOGS
#define GST_LOG_EXTRA(args...)  GST_LOG(args)
#else
#define GST_LOG_EXTRA(args...)
#endif

size_t fd_read(int fd, void *buf, size_t n);
size_t fd_write(int fd, void *buf, size_t n);

/* v4l2 notifications */
enum { 
	CAMSRC_N_INVALID = 0,
	CAMSRC_N_STILL_CAPTURE,
	CAMSRC_N_BUFFER_READY,
	CAMSRC_N_FLASH_MODE,
	CAMSRC_N_PAUSED,
};

#define CAMSRC_W_NOTIFICATION(x) \
	do { \
		int msg = x; \
		int fd = v4l2src->v4l2object->propset_notify_fd[1]; \
		if (fd != -1) {\
			g_mutex_lock(v4l2src->notification_lock); \
			gint size = fd_write(fd, &msg, sizeof(msg)); \
			g_assert(size == sizeof(msg)); \
			g_mutex_unlock(v4l2src->notification_lock); \
		} \
	} while (0)

#define CAMSRC_R_NOTIFICATION(msg_ref) \
	do { \
		g_mutex_lock(v4l2src->notification_lock); \
		gint size = fd_read(v4l2src->v4l2object->propset_notify_fd[0], \
				msg_ref, sizeof(gint)); \
		g_assert(size == sizeof(gint)); \
		g_mutex_unlock(v4l2src->notification_lock); \
	} while (0)

gboolean send_camd_notification(GstV4l2Src * v4l2src,
		camd_msg_id_t, camd_msg_arg_t a,
		void *ptr, unsigned int size,
		const char *desc);

#define CAMD_NOTIFICATION_SEND(v,e,a) send_camd_notification(v,e,a,NULL,0,#e ":" #a)
#define CAMD_NOTIFICATION_REQ_DATA(v,e,a,p,s) send_camd_notification(v,e,a,p,s,#e ":" #a)

#endif /* __GST_V4L2SRC_H__ */
